from biotransformers.bio_transformers import BioTransformers  # noqa
from biotransformers.utils.logger import logger  # noqa

from .version import VERSION  # noqa

log = logger("biotransformers")

__version__ = VERSION
